{-# OPTIONS -fno-implicit-prelude #-}
{- |
ToDo:
use AffineSpace instead of Module for the particular interpolation types,
since affine combinations assert reconstruction of constant functions.
They are more natural for interpolation of internal control parameters.
However, how can cubic interpolation expressed by affine combinations
without divisions?
-}
module Synthesizer.State.Interpolation where

import qualified Synthesizer.State.Signal  as Sig
import qualified Synthesizer.Plain.Control as Ctrl

import qualified Synthesizer.Generic.Interpolation as InterpolationG
import qualified Synthesizer.Generic.Signal as SigG
import qualified Synthesizer.Generic.SampledValue as Sample

import qualified Algebra.Module    as Module
import qualified Algebra.RealField as RealField
import qualified Algebra.Field     as Field
import qualified Algebra.Ring      as Ring
import qualified Algebra.Additive  as Additive

import Algebra.Module((*>))
import Data.Maybe (fromMaybe)

import Control.Monad.State (StateT(StateT), evalStateT, ap, )
import Control.Applicative (Applicative(pure, (<*>)), (<$>), liftA2, )
import Synthesizer.ApplicativeUtility (liftA4, )
import Synthesizer.Utility (affineComb, )

import PreludeBase
import NumericPrelude




{- | interpolation as needed for resampling -}
data T t y =
  Cons {
    number :: Int,  -- interpolation requires a total number of 'number'
    offset :: Int,  -- interpolation requires 'offset' values before the current
    func   :: t -> Sig.T y -> y
  }


{-# INLINE toGeneric #-}
toGeneric ::
   (Sample.C y, SigG.C sig) =>
   T t y -> InterpolationG.T sig t y
toGeneric ip =
   InterpolationG.Cons {
      InterpolationG.number = number ip,
      InterpolationG.offset = offset ip,
      InterpolationG.func = \ t x -> func ip t (Sig.fromGenericSignal x)
   }



{-* Interpolation with various padding methods -}

{-# INLINE zeroPad #-}
zeroPad :: (RealField.C t) =>
   (T t y -> t -> Sig.T y -> a) ->
   y -> T t y -> t -> Sig.T y -> a
zeroPad interpolate z ip phase x =
   let (phInt, phFrac) = splitFraction phase
   in  interpolate ip phFrac
          (delayPad z (offset ip - phInt) (Sig.append x (Sig.repeat z)))

{-# INLINE constantPad #-}
constantPad :: (RealField.C t) =>
   (T t y -> t -> Sig.T y -> a) ->
   T t y -> t -> Sig.T y -> a
constantPad interpolate ip phase x =
   let (phInt, phFrac) = splitFraction phase
       xPad =
          do (xFirst,_) <- Sig.viewL x
             return (delayPad xFirst (offset ip - phInt) (Sig.extendConstant x))
   in  interpolate ip phFrac
          (fromMaybe Sig.empty xPad)


{- |
Only for finite input signals.
-}
{-# INLINE cyclicPad #-}
cyclicPad :: (RealField.C t) =>
   (T t y -> t -> Sig.T y -> a) ->
   T t y -> t -> Sig.T y -> a
cyclicPad interpolate ip phase x =
   let (phInt, phFrac) = splitFraction phase
   in  interpolate ip phFrac
          (Sig.drop (mod (phInt - offset ip) (Sig.length x)) (Sig.cycle x))

{- |
The extrapolation may miss some of the first and some of the last points
-}
{-# INLINE extrapolationPad #-}
extrapolationPad :: (RealField.C t) =>
   (T t y -> t -> Sig.T y -> a) ->
   T t y -> t -> Sig.T y -> a
extrapolationPad interpolate ip phase =
   interpolate ip (phase - fromIntegral (offset ip))
{-
  This example shows pikes, although there shouldn't be any:
   plotList (take 100 $ interpolate (Zero (0::Double)) ipCubic (-0.9::Double) (repeat 0.03) [1,0,1,0.8])
-}


{-* Helper methods for interpolation of multiple nodes -}

{-# INLINE skip #-}
skip :: (RealField.C t) =>
   T t y -> (t, Sig.T y) -> (t, Sig.T y)
skip ip (phase0, x0) =
   let (n, frac) = splitFraction phase0
       (m, x1) = Sig.dropMarginRem (number ip) n x0
   in  (fromIntegral m + frac, x1)

{-# INLINE single #-}
single :: (RealField.C t) =>
   T t y -> t -> Sig.T y -> y
single ip phase0 x0 =
   uncurry (func ip) $ skip ip (phase0, x0)
--   curry (uncurry (func ip) . skip ip)
{-
GNUPlot.plotFunc [] (GNUPlot.linearScale 1000 (0,2)) (\t -> single linear (t::Double) [0,4,1::Double])
-}



{-* Different kinds of interpolation -}

{-** Hard-wired interpolations -}

data PrefixReader y a =
   PrefixReader Int (StateT (Sig.T y) Maybe a)

instance Functor (PrefixReader y) where
   fmap f (PrefixReader count parser) =
      PrefixReader count (fmap f parser)

instance Applicative (PrefixReader y) where
   pure = PrefixReader 0 . return
   (PrefixReader count0 parser0) <*> (PrefixReader count1 parser1) =
       PrefixReader (count0+count1) (parser0 `ap` parser1)

{-# INLINE getNode #-}
getNode :: PrefixReader y y
getNode = PrefixReader 1 (StateT Sig.viewL)

{-# INLINE fromPrefixReader #-}
fromPrefixReader :: String -> Int -> PrefixReader y (t -> y) -> T t y
fromPrefixReader name off (PrefixReader count parser) =
   Cons count off
      (\t xs ->
          maybe
             (error (name ++ " interpolation: not enough nodes"))
             ($t)
             (evalStateT parser xs))

{-| Consider the signal to be piecewise constant. -}
{-# INLINE constant #-}
constant :: T t y
constant =
   fromPrefixReader "constant" 0 (const <$> getNode)

{-| Consider the signal to be piecewise linear. -}
{-# INLINE linear #-}
linear :: (Module.C t y) => T t y
linear =
   fromPrefixReader "linear" 0
      (liftA2
          (\x0 x1 phase -> affineComb phase (x0,x1))
          getNode getNode)

{-| Consider the signal to be piecewise cubic,
    with smooth connections at the nodes.
    It uses a cubic curve which has node values
    x0 at 0 and x1 at 1 and derivatives
    (x1-xm1)/2 and (x2-x0)/2, respectively.
    You can see how it works
    if you evaluate the expression for t=0 and t=1
    as well as the derivative at these points. -}
{-# INLINE cubic #-}
cubic :: (Field.C t, Module.C t y) => T t y
cubic =
   fromPrefixReader "cubic" 1
      (liftA4
         (\xm1 x0 x1 x2 t ->
            let lipm12 = affineComb t (xm1,x2)
                lip01  = affineComb t (x0, x1)
                three  = 3 `asTypeOf` t
            in  lip01 + (t*(t-1)/2) *>
                           (lipm12 + (x0+x1) - three *> lip01))
         getNode getNode getNode getNode)

{-# INLINE cubicAlt #-}
cubicAlt :: (Field.C t, Module.C t y) => T t y
cubicAlt =
   fromPrefixReader "cubicAlt" 1
      (liftA4
         (\xm1 x0 x1 x2 t ->
          let half = 1/2 `asTypeOf` t
          in  cubicHalf    t  x0 (half *> (x1-xm1)) +
              cubicHalf (1-t) x1 (half *> (x0-x2)))
         getNode getNode getNode getNode)


{- \t -> cubicHalf t x x' has a double zero at 1 and
   at 0 it has value x and steepness x' -}
{-# INLINE cubicHalf #-}
cubicHalf :: (Module.C t y) => t -> y -> y -> y
cubicHalf t x x' =
   (t-1)^2 *> ((1+2*t)*>x + t*>x')



{-** Interpolation based on piecewise defined functions -}

{-# INLINE piecewise #-}
piecewise :: (Module.C t y) =>
   Int -> [t -> t] -> T t y
piecewise center ps =
   Cons (length ps) (center-1)
      (\t -> Sig.linearComb (Sig.fromList (map ($t) (reverse ps))))

{-# INLINE piecewiseConstant #-}
piecewiseConstant :: (Module.C t y) => T t y
piecewiseConstant =
   piecewise 1 [const 1]

{-# INLINE piecewiseLinear #-}
piecewiseLinear :: (Module.C t y) => T t y
piecewiseLinear =
   piecewise 1 [id, (1-)]

{-# INLINE piecewiseCubic #-}
piecewiseCubic :: (Field.C t, Module.C t y) => T t y
piecewiseCubic =
   piecewise 2 $
      Ctrl.cubicFunc (0,(0,0))    (1,(0,1/2)) :
      Ctrl.cubicFunc (0,(0,1/2))  (1,(1,0)) :
      Ctrl.cubicFunc (0,(1,0))    (1,(0,-1/2)) :
      Ctrl.cubicFunc (0,(0,-1/2)) (1,(0,0)) :
      []

{-
GNUPlot.plotList [] $ take 100 $ interpolate (Zero 0) piecewiseCubic (-2.3 :: Double) (repeat 0.1) [2,1,2::Double]
-}


{-** Interpolation based on arbitrary functions -}

{- | with this wrapper you can use the collection of interpolating functions from Donadio's DSP library -}
{-# INLINE function #-}
function :: (Module.C t y) =>
      (Int,Int)   {- ^ @(left extent, right extent)@, e.g. @(1,1)@ for linear hat -}
   -> (t -> t)
   -> T t y
function (left,right) f =
   let len = left+right
       ps  = Sig.take len $ Sig.iterate pred (pred right)
       -- ps = Sig.reverse $ Sig.take len $ Sig.iterate succ (-left)
   in  Cons len left
          (\t -> Sig.linearComb $
                   Sig.map (\x -> f (t + fromIntegral x)) ps)
{-
GNUPlot.plotList [] $ take 300 $ interpolate (Zero 0) (function (1,1) (\x -> exp (-6*x*x))) (-2.3 :: Double) (repeat 0.03) [2,1,2::Double]
-}


{-* Helper functions -}

{-# INLINE delayPad #-}
delayPad :: y -> Int -> Sig.T y -> Sig.T y
delayPad z n =
   if n<0
     then Sig.drop (negate n)
     else Sig.append (Sig.replicate n z)
